@@ -88,7 +88,7 @@ ArrayViewport <- function(refdim, ranges=NULL)
 
 make_string_from_ArrayViewport <- function(viewport, dimnames=NULL,
                                            as.2Dslice=FALSE,
-                                           with.brackets=FALSE)
+                                           collapse=",", with.brackets=FALSE)
 {
     if (!isTRUEorFALSE(as.2Dslice))
         stop("'as.2Dslice' must be TRUE or FALSE")
@@ -120,7 +120,7 @@ make_string_from_ArrayViewport <- function(viewport, dimnames=NULL,
     }
     if (ans[[1L]] == "" && with.brackets)
         ans[[1L]] <- " "
-    ans <- paste0(ans, collapse=", ")
+    ans <- paste0(ans, collapse=collapse)
     if (with.brackets)
         ans <- paste0("[", ans, "]")
     ans
@@ -129,9 +129,10 @@ make_string_from_ArrayViewport <- function(viewport, dimnames=NULL,
 setMethod("show", "ArrayViewport",
     function(object)
     {
+        dim_in1string <- paste0(dim(object), collapse=" x ")
         refdim_in1string <- paste0(refdim(object), collapse=" x ")
-        cat(class(object), " object on a ", refdim_in1string, " array: ",
-            sep="")
+        cat(dim_in1string, " ", class(object), " object on a ",
+            refdim_in1string, " array: ", sep="")
         s <- make_string_from_ArrayViewport(object, with.brackets=TRUE)
         cat(s, "\n", sep="")
     }
@@ -146,7 +147,7 @@ makeNindexFromArrayViewport <- function(viewport, expand.RangeNSBS=FALSE)
     viewport_dim <- dim(viewport)
     viewport_refdim <- refdim(viewport)
     ndim <- length(viewport_dim)
-    Nindex <- vector(mode="list", length=ndim)
+    Nindex <- vector("list", length=ndim)
     is_not_missing <- viewport_dim < viewport_refdim
     if (expand.RangeNSBS) {
         expand_idx <- which(is_not_missing)
@@ -167,30 +168,12 @@ makeNindexFromArrayViewport <- function(viewport, expand.RangeNSBS=FALSE)
             }
         )
     }
-    Nindex[expand_idx] <- as.list(as(viewport_ranges[expand_idx],
-                                     "CompressedIntegerList"))
-
+    if (length(expand_idx) != 0L)
+        Nindex[expand_idx] <- as.list(as(viewport_ranges[expand_idx],
+                                         "CompressedIntegerList"))
     Nindex
 }
 
-### 2 utilities for extracting/replacing blocks from/in an array-like object
-
-extract_block <- function(x, viewport)
-{
-    stopifnot(identical(dim(x), refdim(viewport)))
-    Nindex <- makeNindexFromArrayViewport(viewport)
-    subset_by_Nindex(x, Nindex)
-}
-
-### Return the modified array.
-replace_block <- function(x, viewport, block)
-{
-    stopifnot(identical(dim(x), refdim(viewport)),
-              identical(dim(viewport), dim(block)))
-    Nindex <- makeNindexFromArrayViewport(viewport)
-    replace_by_Nindex(x, Nindex, block)
-}
-
 
 ### - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 ### ArrayGrid objects
@@ -253,12 +236,32 @@ setClass("RegularArrayGrid",
 .get_ArbitraryArrayGrid_max_spacings <- function(x)
 {
     vapply(seq_along(x@tickmarks),
-           function(along)
-               max(0L, .get_ArbitraryArrayGrid_spacings_along(x, along)),
-           integer(1))
+        function(along)
+            max(0L, .get_ArbitraryArrayGrid_spacings_along(x, along)),
+        integer(1),
+        USE.NAMES=FALSE
+    )
 }
 
-.get_RegularArrayGrid_dim <- function(refdim, spacings)
+### Get length of biggest viewport in ArbitraryArrayGrid object 'x'.
+.get_ArbitraryArrayGrid_maxlength <- function(x)
+{
+    ans <- prod(.get_ArbitraryArrayGrid_max_spacings(x))
+    if (ans <= .Machine$integer.max)
+        ans <- as.integer(ans)
+    ans
+}
+
+### Get length of biggest viewport in RegularArrayGrid object 'x'.
+.get_RegularArrayGrid_maxlength <- function(x)
+{
+    ans <- prod(x@spacings)
+    if (ans <= .Machine$integer.max)
+        ans <- as.integer(ans)
+    ans
+}
+
+get_RegularArrayGrid_dim <- function(refdim, spacings)
 {
     ans <- refdim %/% spacings + (refdim %% spacings != 0L)
     ans[is.na(ans)] <- 1L
@@ -293,8 +296,8 @@ setClass("RegularArrayGrid",
     if (!all(ok))
         return(wmsg2("each list element in 'tickmarks' slot must be a ",
                      "sorted integer vector of non-negative values"))
-    max_spacings <- .get_ArbitraryArrayGrid_max_spacings(x)
-    if (prod(max_spacings) > .Machine$integer.max)
+    x_maxlen <- .get_ArbitraryArrayGrid_maxlength(x)
+    if (x_maxlen > .Machine$integer.max)
         return(wmsg2("grid is too coarse (all grid elements must have a ",
                      "length <= .Machine$integer.max)"))
     TRUE
@@ -315,12 +318,13 @@ setValidity2("ArbitraryArrayGrid", .validate_ArbitraryArrayGrid)
         return(wmsg2("'spacings' and 'refdim' slots must have ",
                      "the same length"))
     if (!all(x_spacings <= x_refdim))
-        return(wmsg2("values in 'spacings' slot must be <= the ",
-                     "corresponding values in 'refdim' slot"))
+        return(wmsg2("values in 'spacings' slot must be <= their ",
+                     "corresponding value in 'refdim' slot"))
     if (any(x_spacings == 0L & x_refdim != 0L))
-        return(wmsg2("values in 'spacings' slot cannot be 0 unless the ",
-                     "corresponding values in 'refdim' slot are 0"))
-    if (prod(x_spacings) > .Machine$integer.max)
+        return(wmsg2("values in 'spacings' slot cannot be 0 unless their ",
+                     "corresponding value in 'refdim' slot is also 0"))
+    x_maxlen <- .get_RegularArrayGrid_maxlength(x)
+    if (x_maxlen > .Machine$integer.max)
         return(wmsg2("grid is too coarse (all grid elements must have a ",
                      "length <= .Machine$integer.max)"))
     TRUE
@@ -332,10 +336,12 @@ setValidity2("RegularArrayGrid", .validate_RegularArrayGrid)
 setMethod("refdim", "ArbitraryArrayGrid",
     function(x)
     {
-        mapply(function(tm, tm_len) if (tm_len == 0L) 0L else tm[[tm_len]],
-               x@tickmarks,
-               lengths(x@tickmarks),
-               USE.NAMES=FALSE)
+        mapply(
+            function(tm, tm_len) if (tm_len == 0L) 0L else tm[[tm_len]],
+            x@tickmarks,
+            lengths(x@tickmarks),
+            USE.NAMES=FALSE
+        )
     }
 )
 
@@ -344,7 +350,7 @@ setMethod("refdim", "RegularArrayGrid", function(x) x@refdim)
 setMethod("dim", "ArbitraryArrayGrid", function(x) lengths(x@tickmarks))
 
 setMethod("dim", "RegularArrayGrid",
-    function(x) .get_RegularArrayGrid_dim(refdim(x), x@spacings)
+    function(x) get_RegularArrayGrid_dim(refdim(x), x@spacings)
 )
 
 ### Constructors
@@ -386,12 +392,14 @@ setMethod("getArrayElement", "ArbitraryArrayGrid",
     function(x, subscripts)
     {
         x_refdim <- refdim(x)
-        ans_end <- mapply(`[[`, x@tickmarks, subscripts)
+        ans_end <- mapply(`[[`, x@tickmarks, subscripts, USE.NAMES=FALSE)
         ans_width <- mapply(
             function(along, i)
                 .get_ArbitraryArrayGrid_spacings_along(x, along)[[i]],
             seq_along(x_refdim),
-            subscripts)
+            subscripts,
+            USE.NAMES=FALSE
+        )
         ans_ranges <- IRanges(end=ans_end, width=ans_width)
         ArrayViewport(x_refdim, ans_ranges)
     }
@@ -408,7 +416,7 @@ setMethod("getArrayElement", "RegularArrayGrid",
     }
 )
 
-### lengths()
+### dims() and lengths()
 
 ### NOT exported.
 setGeneric("get_spacings_along", signature="x",
@@ -421,33 +429,65 @@ setMethod("get_spacings_along", "RegularArrayGrid",
     .get_RegularArrayGrid_spacings_along
 )
 
-### Equivalent to 'vapply(x, length, integer(1))' but faster.
-### The sum of the hyper-volumes of all the grid elements should be equal
-### to the hyper-volume of the reference array.
-### More concisely: sum(lengths(x)) should be equal to 'prod(refdim(x))'.
-setMethod("lengths", "ArrayGrid",
-    function (x, use.names=TRUE)
+### Equivalent to 't(vapply(x, dim, refdim(x)))' but faster.
+setMethod("dims", "ArrayGrid",
+    function(x)
     {
-        ans <- get_spacings_along(x, 1L)
+        ans <- as.matrix(get_spacings_along(x, 1L))
         x_ndim <- length(refdim(x))
         if (x_ndim >= 2L) {
-            for (along in 2:x_ndim)
-              ans <- ans * rep(get_spacings_along(x, along), each=length(ans))
+            for (along in 2:x_ndim) {
+                spacings_along <- get_spacings_along(x, along)
+                ans <- cbind(
+                    S4Vectors:::rep.int_along_ROWS(ans, length(spacings_along)),
+                    rep(spacings_along, each=nrow(ans))
+                )
+            }
         }
         ans
     }
 )
 
+### Equivalent to 'vapply(x, length, integer(1))' or to 'rowProds(dims(x))'
+### but faster.
+### The sum of the hyper-volumes of all the grid elements should be equal
+### to the hyper-volume of the reference array.
+### More concisely: sum(lengths(x)) should be equal to 'prod(refdim(x))'.
+setMethod("lengths", "ArrayGrid",
+    function(x, use.names=TRUE)
+    {
+        ans <- get_spacings_along(x, 1L)
+        x_ndim <- length(refdim(x))
+        if (x_ndim >= 2L) {
+            for (along in 2:x_ndim) {
+                spacings_along <- get_spacings_along(x, along)
+                ans <- ans * rep(spacings_along, each=length(ans))
+            }
+        }
+        ans
+    }
+)
+
+### Equivalent to 'max(lengths(x))' except that when 'x' is an ArrayGrid
+### object this can be computed without computing 'lengths(x)' first so is
+### very efficient.
+setGeneric("maxlength", function(x) standardGeneric("maxlength"))
+setMethod("maxlength", "ANY", function(x) max(lengths(x)))
+setMethod("maxlength", "ArbitraryArrayGrid", .get_ArbitraryArrayGrid_maxlength)
+setMethod("maxlength", "RegularArrayGrid", .get_RegularArrayGrid_maxlength)
+
 ### Show
 
 ### S3/S4 combo for as.character.ArrayGrid
-.as.character.ArrayGrid <- function(x, with.brackets=FALSE)
+.as.character.ArrayGrid <- function(x, collapse=",", with.brackets=FALSE)
 {
     data <- vapply(x,
         function(viewport)
             make_string_from_ArrayViewport(viewport,
+                                           collapse=collapse,
                                            with.brackets=with.brackets),
-        character(1)
+        character(1),
+        USE.NAMES=FALSE
     )
     array(data, dim(x))
 }
@@ -459,181 +499,114 @@ setMethod("show", "ArrayGrid",
     {
         dim_in1string <- paste0(dim(object), collapse=" x ")
         refdim_in1string <- paste0(refdim(object), collapse=" x ")
-        cat(dim_in1string, " ", class(object), " object ",
-            "on a ", refdim_in1string, " array:\n", sep="")
+        cat(dim_in1string, " ", class(object), " object on a ",
+            refdim_in1string, " array:\n", sep="")
         ## Turn 'object' into a character array.
-        print(as.character(object, TRUE), quote=FALSE)
+        print(as.character(object, with.brackets=TRUE),
+              quote=FALSE, right=TRUE)
     }
 )
 
 
 ### - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
-### get_spacings_for_hypercube_capped_length_blocks()
+### aperm()
 ###
-### Typically used to create a regular grid with a first block that is as
-### close as possible to an hypercube and that has a length as close as
-### possibe to (but not bigger than) 'max_block_len'.
-### NOT exported but used in HDF5Array!
-get_spacings_for_hypercube_capped_length_blocks <-
-    function(refdim, max_block_len)
+### Extend base::aperm() by allowing dropping and/or adding ineffective
+### dimensions. See aperm2.R
+###
+
+.aperm.ArbitraryArrayGrid <- function(a, perm)
 {
-    if (!isSingleNumber(max_block_len))
-        stop("'max_block_len' must be a single integer")
-    if (!is.integer(max_block_len))
-        max_block_len <- as.integer(max_block_len)
-
-    p <- prod(refdim)
-    if (p <= max_block_len)
-        return(refdim)
-
-    spacings <- refdim
-    L <- max(spacings)
-    while (TRUE) {
-        is_max <- spacings == L
-        not_max_spacings <- spacings[!is_max]
-        L <- (max_block_len / prod(not_max_spacings)) ^ (1 / sum(is_max))
-        if (length(not_max_spacings) == 0L)
-            break
-        L2 <- max(not_max_spacings)
-        if (L >= L2)
-            break
-        L <- L2
-        spacings[is_max] <- L
-    }
-    spacings[is_max] <- as.integer(L)
-    q <- .get_RegularArrayGrid_dim(refdim, spacings + 1L) /
-         .get_RegularArrayGrid_dim(refdim, spacings)
-    for (along in which(is_max)[order(q[is_max])]) {
-        spacings[[along]] <- spacings[[along]] + 1L
-        p <- prod(spacings)
-        if (p == max_block_len)
-            break
-        if (p > max_block_len) {
-            spacings[[along]] <- spacings[[along]] - 1L
-            break
-        }
-    }
-    spacings
+    ## We don't perform a full check of 'perm' (see normarg_perm() in
+    ## aperm2.R) because we want to allow duplicates in it. Instead we
+    ## call replaceSlots() with 'check=TRUE' to trigger validation of
+    ## the modified ArrayGrid object with the expectation that it will
+    ## fail if for example 'perm' contains NAs or values > length(dim(a)).
+    perm <- normarg_perm(perm, dim(a))
+    ans_tickmarks <- a@tickmarks[perm]
+    ans_tickmarks[is.na(perm)] <- list(1L)
+    BiocGenerics:::replaceSlots(a, tickmarks=ans_tickmarks, check=TRUE)
 }
+### S3/S4 combo for aperm.ArbitraryArrayGrid
+aperm.ArbitraryArrayGrid <-
+    function(a, perm, ...) .aperm.ArbitraryArrayGrid(a, perm, ...)
+setMethod("aperm", "ArbitraryArrayGrid", aperm.ArbitraryArrayGrid)
+
+.aperm.RegularArrayGrid <- function(a, perm)
+{
+    ## See above for why it's important to call replaceSlots() with
+    ## 'check=TRUE'.
+    perm <- normarg_perm(perm, dim(a))
+    ans_refdim <- a@refdim[perm]
+    ans_refdim[is.na(perm)] <- 1L
+    ans_spacings <- a@spacings[perm]
+    ans_spacings[is.na(perm)] <- 1L
+    BiocGenerics:::replaceSlots(a, refdim=ans_refdim,
+                                   spacings=ans_spacings,
+                                   check=TRUE)
+}
+### S3/S4 combo for aperm.RegularArrayGrid
+aperm.RegularArrayGrid <-
+    function(a, perm, ...) .aperm.RegularArrayGrid(a, perm, ...)
+setMethod("aperm", "RegularArrayGrid", aperm.RegularArrayGrid)
 
 
 ### - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
-### Linear blocks
+### downsample() generic and methods
 ###
-### An array block is "linear" if it would be made of elements contiguous in
-### memory if the reference array was an ordinary R array (where the fastest
-### changing dimension is the first one).
-### A grid element is "linear" if it defines a linear block.
+### Reduce the "resolution" of a grid by the specified ratio.
+### Act as an endomorphism.
 ###
 
-setGeneric("isLinear", function(x) standardGeneric("isLinear"))
+setGeneric("downsample", signature="x",
+    function(x, ratio=1L) standardGeneric("downsample")
+)
 
-setMethod("isLinear", "ArrayViewport",
-    function(x)
+.normarg_ratio <- function(ratio, x_dim)
+{
+    if (!is.numeric(ratio))
+        stop(wmsg("'ratio' must be an integer vector"))
+    if (!is.integer(ratio))
+        ratio <- as.integer(ratio)
+    ndim <- length(x_dim)
+    if (length(ratio) != 1L && length(ratio) != ndim)
+        stop(wmsg("'length(ratio)' must be 1 or the sane as 'length(dim(x))'"))
+    if (S4Vectors:::anyMissingOrOutside(ratio, 0L))
+        stop(wmsg("'ratio' cannot contain negative or NA values"))
+    if (length(ratio) != ndim)
+        ratio <- rep.int(ratio, ndim)
+    if (any(ratio == 0L & x_dim != 0L))
+        stop(wmsg("values in 'ratio' cannot be 0 unless their ",
+                  "corresponding dimension in 'x' is also 0"))
+    ratio
+}
+
+setMethod("downsample", "ArbitraryArrayGrid",
+    function(x, ratio=1L)
     {
-        x_width <- width(x)
-        idx <- which(x_width != refdim(x))
-        if (length(idx) == 0L)
-            return(TRUE)
-        all(tail(x_width, n=-idx[[1L]]) == 1L)
+        ratio <- .normarg_ratio(ratio, dim(x))
+        ans_tickmarks <- mapply(
+            function(tm, tm_len, r) {
+                if (tm_len == 0L)
+                    return(integer(0))
+                tm[seq2(tm_len, r)]
+            },
+            x@tickmarks,
+            lengths(x@tickmarks),
+            ratio,
+            SIMPLIFY=FALSE,
+            USE.NAMES=FALSE
+        )
+        ArbitraryArrayGrid(ans_tickmarks)
     }
 )
 
-### If the 1st grid element is linear, then they all are.
-setMethod("isLinear", "ArrayGrid",
-    function(x)
+setMethod("downsample", "RegularArrayGrid",
+    function(x, ratio=1L)
     {
-        if (length(x) == 0L)
-            return(TRUE)
-        isLinear(x[[1L]])
+        ratio <- .normarg_ratio(ratio, dim(x))
+        ans_spacings <- pmin(x@spacings * ratio, refdim(x))
+        RegularArrayGrid(refdim(x), ans_spacings)
     }
 )
 
-### Typically used to create a regular grid with linear blocks of length as
-### close as possibe to (but not bigger than) 'max_block_len'.
-### NOT exported but used in HDF5Array!
-get_spacings_for_linear_capped_length_blocks <- function(refdim, max_block_len)
-{
-    if (!isSingleNumber(max_block_len))
-        stop("'max_block_len' must be a single integer")
-    if (!is.integer(max_block_len))
-        max_block_len <- as.integer(max_block_len)
-
-    p <- cumprod(refdim)
-    w <- which(p <= max_block_len)
-    N <- if (length(w) == 0L) 1L else w[[length(w)]] + 1L
-    if (N > length(refdim))
-        return(refdim)
-    if (N == 1L) {
-        by <- max_block_len
-    } else {
-        by <- max_block_len %/% as.integer(p[[N - 1L]])
-    }
-    c(head(refdim, n=N-1L), by, rep.int(1L, length(refdim)-N))
-}
-
-
-### - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
-### Some unexported utilities
-###
-
-### Used in the DelayedMatrixStats package!
-get_spacings_for_capped_length_blocks <-
-    function(refdim, max_block_len, block_shape=c("hypercube", "linear"))
-{
-    block_shape <- match.arg(block_shape)
-    FUN <- switch(block_shape,
-                  hypercube=get_spacings_for_hypercube_capped_length_blocks,
-                  linear=get_spacings_for_linear_capped_length_blocks,
-                  stop("unsupported 'block_shape'"))
-    FUN(refdim, max_block_len)
-}
-
-make_RegularArrayGrid_of_capped_length_blocks <-
-    function(refdim, max_block_len, block_shape=c("hypercube", "linear"))
-{
-    spacings <- get_spacings_for_capped_length_blocks(
-                    refdim, max_block_len, block_shape=block_shape)
-    RegularArrayGrid(refdim, spacings)
-}
-
-### Used in unit tests.
-split_array_in_capped_length_blocks <-
-    function(x, max_block_len, block_shape=c("hypercube", "linear"))
-{
-    grid <- make_RegularArrayGrid_of_capped_length_blocks(
-                        dim(x), max_block_len, block_shape=block_shape)
-    lapply(grid, function(viewport) extract_block(x, viewport))
-}
-
-### Used in unit tests.
-### Rebuild the original array from the blocks obtained with
-### split_array_in_capped_length_blocks( , block_shape="hypercube") as
-### an *ordinary* array. So if 'x' is an ordinary array, then:
-###
-###   blocks <- split_array_in_capped_length_blocks(x, max_block_len,
-###                                                 block_shape="hypercube")
-###   unsplit_array_from_hypercube_blocks(blocks, x)
-###
-### should be a no-op for any 'max_block_len' <= 'length(x)'.
-unsplit_array_from_hypercube_blocks <- function(blocks, x)
-{
-    stop("Not ready yet")
-}
-
-### Used in unit tests.
-### Rebuild the original array from the blocks obtained with
-### split_array_in_capped_length_blocks( , block_shape="linear") as
-### an *ordinary* array. So if 'x' is an ordinary array, then:
-###
-###   blocks <- split_array_in_capped_length_blocks(x, max_block_len,
-###                                                 block_shape="linear")
-###   unsplit_array_from_linear_blocks(blocks, x)
-###
-### should be a no-op for any 'max_block_len' <= 'length(x)'.
-unsplit_array_from_linear_blocks <- function(blocks, x)
-{
-    ans <- combine_array_objects(blocks)
-    set_dim(ans, dim(x))
-}
-
