@@ -4,7 +4,7 @@
 
 
 ### The "root" node of the tree of DelayedOp objects. Represents a no-op.
-setClass("DelayedArray", contains="DelayedUnaryOp")
+setClass("DelayedArray", contains="DelayedUnaryIsoOp")
 
 ### Extending DataTable gives us a few things for free (head(), tail(),
 ### etc...). Note that even though DelayedMatrix already extends Array (via
@@ -12,16 +12,18 @@ setClass("DelayedArray", contains="DelayedUnaryOp")
 ### a *direct* child of Array and to place Array *before* DataTable in the
 ### 'contains' field below. This ensures that method dispatch will pick the
 ### method for Array in case a generic has methods defined for Array and
-### DataTable (e.g. as.data.frame()). Furthermore, it seems that we also need
-### to place all the classes that are in the inheritance path between
-### DelayedArray and Array in the 'contains' field otherwise we get the
-### following error when trying to instantiate a DelayedMatrix object with
-### new("DelayedMatrix"):
+### DataTable (e.g. as.data.frame()). Furthermore, for some obscure reason,
+### it seems that we also need to place all the classes that are in the
+### inheritance path between DelayedArray and Array in the 'contains' field
+### otherwise we get the following error when trying to instantiate a
+### DelayedMatrix object with new("DelayedMatrix"):
 ###
 ###     Error: C stack usage  7971652 is too close to the limit
 ###
 setClass("DelayedMatrix",
-    contains=c("DelayedArray", "DelayedUnaryOp", "DelayedOp", "Array",
+    contains=c("DelayedArray",
+               "DelayedUnaryIsoOp", "DelayedUnaryOp", "DelayedOp",
+               "Array",
                "DataTable"),
     prototype=prototype(
         seed=new("matrix")
@@ -138,6 +140,38 @@ setValidity2("DelayedArray", .validate_DelayedArray)
 setValidity2("DelayedMatrix", .validate_DelayedMatrix)
 
 
+### - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
+### Seed contract
+###
+### We overwrite the methods for DelayedUnaryOp objects only to detect
+### DelayedArray objects with old internals.
+###
+
+setMethod("dim", "DelayedArray",
+    function(x)
+    {
+        .check_DelayedArray_internals(x)
+        callNextMethod()
+    }
+)
+
+setMethod("dimnames", "DelayedArray",
+    function(x)
+    {
+        .check_DelayedArray_internals(x)
+        callNextMethod()
+    }
+)
+
+setMethod("extract_array", "DelayedArray",
+    function(x, index)
+    {
+        .check_DelayedArray_internals(x)
+        callNextMethod()
+    }
+)
+
+
 ### - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 ### Constructor
 ###
@@ -145,24 +179,72 @@ setValidity2("DelayedMatrix", .validate_DelayedMatrix)
 ### Low-level constructor. Not intended to be used directly by the end user.
 new_DelayedArray <- function(seed=new("array"), Class="DelayedArray")
 {
+    stopifnot(isSingleString(Class), extends(Class, "DelayedArray"))
     seed_ndim <- length(dim(seed))
-    if (seed_ndim == 2L)
-        Class <- matrixClass(new(Class))
+    if (extends(Class, "DelayedMatrix")) {
+        if (seed_ndim != 2L)
+            stop(wmsg("the supplied seed must have exactly 2 dimensions ",
+                      "when the specified class (", Class, ") extends ",
+                      "DelayedMatrix"))
+    } else {
+        if (seed_ndim == 2L)
+            Class <- matrixClass(new(Class))
+    }
     new2(Class, seed=seed)
 }
 
 setGeneric("DelayedArray", function(seed) standardGeneric("DelayedArray"))
 
-setMethod("DelayedArray", "ANY",
+setMethod("DelayedArray", "ANY", function(seed) new_DelayedArray(seed))
+
+setMethod("DelayedArray", "DelayedArray", function(seed) seed)
+
+setMethod("DelayedArray", "DelayedOp",
     function(seed)
     {
-        if (getOption("DelayedArray.simplify", default=TRUE))
+        if (getOption("DelayedArray.simplify", default=TRUE)) {
             seed <- simplify(seed, incremental=TRUE)
+            if (!is(seed, "DelayedOp"))
+                return(DelayedArray(seed))
+        }
         new_DelayedArray(seed)
     }
 )
 
-setMethod("DelayedArray", "DelayedArray", function(seed) seed)
+
+### - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
+### Pristine objects
+###
+### A pristine DelayedArray object is an object that does not carry any
+### delayed operation.
+###
+
+### Note that false negatives happen when 'x' carries delayed operations that
+### do nothing, but that's ok.
+is_pristine <- function(x) { !is(x@seed, "DelayedOp") }
+
+
+### - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
+### simplify()
+###
+
+setMethod("simplify", "DelayedArray",
+    function(x, incremental=FALSE)
+        DelayedArray(simplify(x@seed, incremental=incremental))
+)
+
+
+### - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
+### netSubsetAndAperm()
+###
+
+setMethod("netSubsetAndAperm", "DelayedArray",
+    function(x, as.DelayedOp=FALSE)
+    {
+        x <- x@seed
+        callGeneric()
+    }
+)
 
 
 ### - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
@@ -200,7 +282,7 @@ setClass("DelayedArray1",
 
     index <- lapply(unname(object1@index), unname)
     op <- new2("DelayedSubset", seed=seed, index=index)
-    if (!isNoOp(op))
+    if (!is_noop(op))
         seed <- op
 
     object1_dimnames <- lapply(seq_along(object1@index),
@@ -212,11 +294,11 @@ setClass("DelayedArray1",
         object1_dimnames <- NULL
 
     op <- new_DelayedDimnames(seed, object1_dimnames)
-    if (!isNoOp(op))
+    if (!is_noop(op))
         seed <- op
 
-    ## Translate 'delayed_ops' slot as DelayedUnaryIsoOp objects and
-    ## stash them inside 'seed'.
+    ## Translate 'delayed_ops' slot as DelayedUnaryIsoOpWithArgs objects
+    ## and stash them inside 'seed'.
 
     for (delayed_op in object1@delayed_ops) {
         OP <- delayed_op[[1L]]
@@ -227,8 +309,8 @@ setClass("DelayedArray1",
         if (recycle_along_last_dim) {
             if (length(Largs) == 1L) Lalong <- 1L else Ralong <- 1L
         }
-        seed <- new_DelayedUnaryIsoOp(seed, OP, Largs, Rargs,
-                                            Lalong, Ralong)
+        seed <- new_DelayedUnaryIsoOpWithArgs(seed, OP, Largs, Rargs,
+                                              Lalong, Ralong)
     }
 
     DelayedArray(seed)
@@ -309,61 +391,6 @@ setClass("DelayedArray1",
 setMethod("updateObject", "DelayedArray", .updateObject_DelayedArray)
 
 
-### - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
-### Seed contract
-###
-### We overwrite the methods for DelayedUnaryOp objects only to detect
-### DelayedArray objects with old internals.
-###
-
-setMethod("dim", "DelayedArray",
-    function(x)
-    {
-        .check_DelayedArray_internals(x)
-        callNextMethod()
-    }
-)
-
-setMethod("dimnames", "DelayedArray",
-    function(x)
-    {
-        .check_DelayedArray_internals(x)
-        callNextMethod()
-    }
-)
-
-setMethod("extract_array", "DelayedArray",
-    function(x, index)
-    {
-        .check_DelayedArray_internals(x)
-        callNextMethod()
-    }
-)
-
-
-### - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
-### simplify()
-###
-
-setMethod("simplify", "DelayedArray",
-    function(x, incremental=FALSE)
-        DelayedArray(simplify(x@seed, incremental=incremental))
-)
-
-
-### - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
-### netSubsetAndAperm()
-###
-
-setMethod("netSubsetAndAperm", "DelayedArray",
-    function(x, as.DelayedOp=FALSE)
-    {
-        x <- x@seed
-        callGeneric()
-    }
-)
-
-
 ### - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 ### Stash delayed ops in a DelayedArray object
 ###
@@ -380,55 +407,91 @@ setMethod("netSubsetAndAperm", "DelayedArray",
 
 stash_DelayedSubset <- function(x, Nindex)
 {
+    stopifnot(is(x, "DelayedArray"))
     op <- new_DelayedSubset(x@seed, Nindex)
     DelayedArray(op)
 }
 
 stash_DelayedAperm <- function(x, perm)
 {
+    stopifnot(is(x, "DelayedArray"))
     op <- new_DelayedAperm(x@seed, perm)
     DelayedArray(op)
 }
 
-stash_DelayedUnaryIsoOp <- function(x, OP, Largs=list(), Rargs=list(),
-                                       Lalong=NA, Ralong=NA)
+stash_DelayedUnaryIsoOpStack <- function(x, OP)
 {
-    op <- new_DelayedUnaryIsoOp(x@seed, OP=OP, Largs=Largs, Rargs=Rargs,
-                                        Lalong=Lalong, Ralong=Ralong)
+    stopifnot(is(x, "DelayedArray"))
+    op <- new_DelayedUnaryIsoOpStack(x@seed, OPS=list(OP), check.op=TRUE)
+    DelayedArray(op)
+}
+
+stash_DelayedUnaryIsoOpWithArgs <- function(x, OP,
+                                            Largs=list(), Rargs=list(),
+                                            Lalong=NA, Ralong=NA)
+{
+    stopifnot(is(x, "DelayedArray"))
+    op <- new_DelayedUnaryIsoOpWithArgs(x@seed, OP=OP,
+                                        Largs=Largs, Rargs=Rargs,
+                                        Lalong=Lalong, Ralong=Ralong,
+                                        check.op=TRUE)
+    DelayedArray(op)
+}
+
+### stash_DelayedUnaryIsoOp() was introduced in BioC 3.7 and renamed
+### stash_DelayedUnaryIsoOpWithArgs() in BioC 3.8. The alias below is for
+### backward compatibility with objects that were serialized with BioC 3.7.
+### For example, the 'BS.cancer.ex.fit' dataset (a BSseq instance from the
+### bsseqData package and used in the bsseq_analysis.Rmd vignette from the
+### bsseq package) contains the "plogis" method for DelayedArray objects in
+### its "trans" slot, and this method contains a reference to
+### stash_DelayedUnaryIsoOp. This is the sign that this BSseq instance was
+### serialized with BioC 3.7. Without the alias below, the bsseq_analysis.Rmd
+### vignette chokes on this line:
+###
+###   BS.cancer.ex.tstat <- BSmooth.tstat(BS.cancer.ex.fit, ...)
+###
+### with the following error:
+###
+###   Error: processing vignette 'bsseq_analysis.Rmd' failed with diagnostics:
+###   could not find function "stash_DelayedUnaryIsoOp"
+###
+### Note that 'BS.cancer.ex.fit' contains 3 DelayedMatrix objects in
+### its "assays" slot.
+### TODO: Remove this alias at some point (e.g. in BioC 3.10).
+stash_DelayedUnaryIsoOp <- stash_DelayedUnaryIsoOpWithArgs
+
+stash_DelayedSubassign <- function(x, Nindex, value)
+{
+    stopifnot(is(x, "DelayedArray"))
+    if (is(value, "DelayedArray"))
+        value <- value@seed
+    op <- new_DelayedSubassign(x@seed, Nindex, value)
     DelayedArray(op)
 }
 
 stash_DelayedDimnames <- function(x, dimnames)
 {
+    stopifnot(is(x, "DelayedArray"))
     op <- new_DelayedDimnames(x@seed, dimnames)
     DelayedArray(op)
 }
 
-stash_DelayedAbind <- function(..., along)
+stash_DelayedAbind <- function(x, objects, along)
 {
-    seeds <- lapply(unname(list(...)), slot, "seed")
+    stopifnot(is(x, "DelayedArray"))
+    objects <- S4Vectors:::prepare_objects_to_bind(x, objects)
+    seeds <- lapply(c(list(x), objects), slot, "seed")
     op <- new_DelayedAbind(seeds, along)
     DelayedArray(op)
 }
 
 
-### - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
-### Pristine objects
-###
-### A pristine DelayedArray object is an object that does not carry any
-### delayed operation.
-###
-
-### Note that false negatives happen when 'x' carries delayed operations that
-### do nothing, but that's ok.
-is_pristine <- function(x) { !is(x@seed, "DelayedOp") }
-
-
 ### - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 ### nseed()
 ###
-### Return the number of leaves in the tree of DelayedOp objects contained
-### in a DelayedArray object.
+### Return the number of leaves in the tree of DelayedOp nodes represented
+### by 'x'. Note that nseed(x) == 1 means that the tree is linear.
 ###
 
 setGeneric("nseed", function(x) standardGeneric("nseed"))
@@ -452,10 +515,10 @@ setMethod("nseed", "ANY",
 ### - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 ### seed() getter/setter
 ###
+### If nseed(x) == 1 (i.e. if 'x' is a linear tree) then the seed() getter
+### and setter MUST work on 'x'.
+###
 
-### If the tree of DelayedOp objects contained in the DelayedArray object
-### has a single leaf (i.e. if the tree is linear), then seed() returns it.
-### Otherwise, it raises an error.
 setGeneric("seed", function(x) standardGeneric("seed"))
 
 setMethod("seed", "DelayedOp",
@@ -467,15 +530,12 @@ setMethod("seed", "DelayedOp",
         }
         x1 <- x@seed
         if (!is(x1, "DelayedOp"))
-            return(x1)  ## found the leaf seed
+            return(x1)  # found the leaf seed
         x <- x1
         callGeneric()  # recursive call
     }
 )
 
-### If the tree of DelayedOp objects contained in the DelayedArray object
-### has a single leaf (i.e. if the tree is linear), then the seed() setter
-### replaces it. Otherwise, it raises an error.
 setGeneric("seed<-", signature="x",
     function(x, ..., value) standardGeneric("seed<-")
 )
@@ -506,7 +566,8 @@ setReplaceMethod("seed", "DelayedOp",
             x@seed <- .normalize_seed_replacement_value(value, x1)
             return(x)
         }
-        seed(x@seed) <- value  # recursive call
+        seed(x1) <- value  # recursive call
+        x@seed <- x1
         x
     }
 )
@@ -541,41 +602,15 @@ setReplaceMethod("path", "DelayedArray",
 ### - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 ### aperm()
 ###
-### Unlike base::aperm(), the method for DelayedArray objects supports
-### dropping dimensions. Note that only "ineffective" dimensions can be
-### dropped (i.e. dimensions equal to 1, so dropping them preserves the
-### length). This feature is used by the "drop" method below.
+### Extend base::aperm() by allowing dropping and/or adding ineffective
+### dimensions. See aperm2.R
 ###
 
-setGeneric("aperm", signature="a")
-
-.aperm.DelayedArray <- function(a, perm)
-{
-    if (missing(perm))
-        perm <- rev(seq_along(dim(a)))
-    stash_DelayedAperm(a, perm)
-}
-
 ### S3/S4 combo for aperm.DelayedArray
-aperm.DelayedArray <- function(a, perm, ...) .aperm.DelayedArray(a, perm, ...)
+aperm.DelayedArray <- function(a, perm, ...) stash_DelayedAperm(a, perm, ...)
 setMethod("aperm", "DelayedArray", aperm.DelayedArray)
 
 
-### - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
-### drop()
-###
-
-setMethod("drop", "DelayedArray",
-    function(x)
-    {
-        perm <- which(dim(x) != 1L)
-        if (length(perm) == 0L)
-            perm <- 1L
-        aperm(x, perm)
-    }
-)
-
-
 ### - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 ### dim() setter
 ###
@@ -597,8 +632,6 @@ setMethod("drop", "DelayedArray",
     if (S4Vectors:::anyMissingOrOutside(value, 0L))
         stop(wmsg("the supplied dim vector cannot contain negative ",
                   "or NA values"))
-    if (length(value) > length(x_dim))
-        stop(wmsg("too many dimensions supplied"))
     prod1 <- prod(value)
     prod2 <- prod(x_dim)
     if (prod1 != prod2)
@@ -615,30 +648,21 @@ setMethod("drop", "DelayedArray",
     cannot_map_msg <- c(
         "Cannot map the supplied dim vector to the current dimensions of ",
         "the object. On a ", x_class, " object, the dim() setter can only ",
-        "be used to drop some of the \"ineffective\" dimensions (i.e. ",
-        "dimensions equal to 1, so dropping them preserves the length of ",
-        "the object)."
+        "be used to drop and/or add \"ineffective dimensions\" (i.e. ",
+        "dimensions equal to 1) to the object."
     )
-
     can_map <- function() {
         if (length(idx1) != length(idx2))
             return(FALSE)
         if (length(idx1) == 0L)
             return(TRUE)
-        if (!all(new_dim[idx1] == old_dim[idx2]))
-            return(FALSE)
-        tmp <- idx2 - idx1
-        tmp[[1L]] >= 0L && isSorted(tmp)
+        all(new_dim[idx1] == old_dim[idx2])
     }
     if (!can_map())
         stop(wmsg(cannot_map_msg))
 
-    new2old <- seq_along(new_dim) +
-        rep.int(c(0L, idx2 - idx1), diff(c(1L, idx1, length(new_dim) + 1L)))
-
-    if (new2old[[length(new2old)]] > length(old_dim))
-        stop(wmsg(cannot_map_msg))
-
+    new2old <- rep.int(NA_integer_, length(new_dim))
+    new2old[idx1] <- idx2
     new2old
 }
 
@@ -647,8 +671,9 @@ setMethod("drop", "DelayedArray",
     x_dim <- dim(x)
     value <- .normalize_dim_replacement_value(value, x_dim)
     new2old <- .map_new_to_old_dim(value, x_dim, class(x))
-    stopifnot(identical(value, x_dim[new2old]))  # sanity check
-    aperm(x, new2old)
+    ans <- aperm(x, new2old)
+    stopifnot(identical(dim(ans), value))  # sanity check
+    ans
 }
 
 setReplaceMethod("dim", "DelayedArray", .set_DelayedArray_dim)
@@ -688,205 +713,6 @@ setMethod("names", "DelayedArray", .get_DelayedArray_names)
 setReplaceMethod("names", "DelayedArray", .set_DelayedArray_names)
 
 
-### - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
-### [
-###
-
-### Linear single bracket subsetting e.g. x[5].
-### NOT delayed (i.e. return an atomic vector).
-.subset_DelayedArray_by_1Dindex <- function(x, i)
-{
-    x_dim <- dim(x)
-    i_dim <- dim(i)
-    if (identical(x_dim, i_dim) && type(i) == "logical") {
-        i <- which(i)  # calls .DelayedArray_block_which() defined
-                       # in DelayedArray-utils.R
-    } else {
-        i <- normalizeSingleBracketSubscript2(i, length(x))
-    }
-    i_len <- length(i)
-    if (i_len == 0L) {
-        ## x0 <- x[integer(0), ..., integer(0)]
-        index <- rep.int(list(integer(0)), length(x_dim))
-        x0 <- extract_array(x, index)
-        return(as.vector(x0))
-    }
-    if (i_len == 1L)
-        return(.get_DelayedArray_element(x, i))
-
-    ## We want to walk only on the blocks that we actually need to visit so we
-    ## don't use block_APPLY() or family because they walk on all the blocks.
-
-    grid <- defaultGrid(x)
-    nblock <- length(grid)
-
-    breakpoints <- cumsum(lengths(grid))
-    part_idx <- get_part_index(i, breakpoints)
-    split_part_idx <- split_part_index(part_idx, length(breakpoints))
-    block_idx <- which(lengths(split_part_idx) != 0L)  # blocks to visit
-    res <- lapply(block_idx, function(b) {
-            if (get_verbose_block_processing())
-                message("Visiting block ", b, "/", nblock, " ... ",
-                        appendLF=FALSE)
-            block <- extract_block(x, grid[[b]])
-            if (!is.array(block))
-                block <- as.array(block)
-            block_ans <- block[split_part_idx[[b]]]
-            if (get_verbose_block_processing())
-                message("OK")
-            block_ans
-    })
-    unlist(res, use.names=FALSE)[get_rev_index(part_idx)]
-}
-
-.subset_DelayedArray <- function(x, i, j, ..., drop=TRUE)
-{
-    if (missing(x))
-        stop("'x' is missing")
-    if (!isTRUEorFALSE(drop))
-        stop("'drop' must be TRUE or FALSE")
-    Nindex <- extract_Nindex_from_syscall(sys.call(), parent.frame())
-    nsubscript <- length(Nindex)
-    if (nsubscript == 0L)
-        return(x)  # no-op
-    x_ndim <- length(dim(x))
-    if (nsubscript == 1L && (x_ndim != 1L || drop)) {
-        ## Linear single bracket subsetting e.g. x[5].
-        ## If 'x' is mono-dimensional and 'drop' is FALSE, we switch
-        ## to "multi-dimensional single bracket subsetting" which is
-        ## delayed.
-        return(.subset_DelayedArray_by_1Dindex(x, Nindex[[1L]]))
-    }
-    if (nsubscript != x_ndim)
-        stop("incorrect number of subscripts")
-    ## Multi-dimensional single bracket subsetting
-    ##
-    ##     x[i_1, i_2, ..., i_n]
-    ##
-    ## Return an object of the same class as 'x' (endomorphism).
-    ans <- stash_DelayedSubset(x, Nindex)
-    if (drop)
-        ans <- drop(ans)
-    ans
-}
-
-setMethod("[", "DelayedArray", .subset_DelayedArray)
-
-
-### - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
-### [<-
-###
-
-.filling_error_msg <- c(
-    "filling a DelayedArray object 'x' with a value (i.e. 'x[] <- value') ",
-    "is supported only when 'value' is an atomic vector and 'length(value)' ",
-    "is a divisor of 'nrow(x)'"
-)
-
-.subassign_error_msg <- c(
-    "subassignment to a DelayedArray object 'x' (i.e. 'x[i] <- value') is ",
-    "supported only when the subscript 'i' is a logical DelayedArray object ",
-    "with the same dimensions as 'x' and when 'value' is a scalar (i.e. an ",
-    "atomic vector of length 1)"
-)
-
-.fill_DelayedArray_with_value <- function(x, value)
-{
-    if (!(is.vector(value) && is.atomic(value)))
-        stop(wmsg(.filling_error_msg))
-    value_len <- length(value)
-    if (value_len == 1L)
-        return(stash_DelayedUnaryIsoOp(x, `[<-`, Rargs=list(value=value)))
-    x_len <- length(x)
-    if (value_len > x_len)
-        stop(wmsg("'value' is longer than 'x'"))
-    x_nrow <- nrow(x)
-    if (x_nrow != 0L) {
-        if (value_len == 0L || x_nrow %% value_len != 0L)
-            stop(wmsg(.filling_error_msg))
-        value <- rep(value, length.out=x_nrow)
-    }
-    stash_DelayedUnaryIsoOp(x, `[<-`, Rargs=list(value=value), Ralong=1L)
-}
-
-.subassign_DelayedArray <- function(x, i, j, ..., value)
-{
-    if (missing(x))
-        stop("'x' is missing")
-    Nindex <- extract_Nindex_from_syscall(sys.call(), parent.frame())
-    nsubscript <- length(Nindex)
-    if (nsubscript == 0L)
-        return(.fill_DelayedArray_with_value(x, value))
-    if (nsubscript != 1L)
-        stop(wmsg(.subassign_error_msg))
-    i <- Nindex[[1L]]
-    if (!(is(i, "DelayedArray") &&
-          identical(dim(x), dim(i)) &&
-          type(i) == "logical"))
-        stop(wmsg(.subassign_error_msg))
-    if (!(is.vector(value) && is.atomic(value) && length(value) == 1L))
-        stop(wmsg(.subassign_error_msg))
-    DelayedArray(new_DelayedNaryIsoOp(x@seed, i@seed,
-                                      OP=`[<-`, Rargs=list(value=value)))
-}
-
-setReplaceMethod("[", "DelayedArray", .subassign_DelayedArray)
-
-
-### - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
-### Coercion to sparse matrix (requires the Matrix package)
-###
-
-.from_DelayedMatrix_to_dgCMatrix <- function(from)
-{
-    idx <- which(from != 0L)
-    array_ind <- arrayInd(idx, dim(from))
-    i <- array_ind[ , 1L]
-    j <- array_ind[ , 2L]
-    x <- from[idx]
-    Matrix::sparseMatrix(i, j, x=x, dims=dim(from), dimnames=dimnames(from))
-}
-
-setAs("DelayedMatrix", "dgCMatrix", .from_DelayedMatrix_to_dgCMatrix)
-setAs("DelayedMatrix", "sparseMatrix", .from_DelayedMatrix_to_dgCMatrix)
-
-
-### - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
-### [[
-###
-### TODO: Implement a "getArrayElement" method for Array objects (in
-### extract_array.R) that is based on extract_array() e.g it just needs
-### to do something like:
-###
-###     as.vector(extract_array(x, as.list(subscripts)))
-###
-### This will make multi-dimensional and linear [[ work on DelayedArray
-### objects. Then remove the method below and update DelayedArray-class.Rd
-###
-
-.get_DelayedArray_element <- function(x, i)
-{
-    i <- normalizeDoubleBracketSubscript(i, x)
-    index <- as.list(arrayInd(i, dim(x)))
-    as.vector(extract_array(x, index))
-}
-
-### Only support linear subsetting at the moment.
-### TODO: Support multi-dimensional subsetting e.g. x[[5, 15, 2]] or
-### x[["E", 15, "b"]].
-setMethod("[[", "DelayedArray",
-    function(x, i, j, ...)
-    {
-        dots <- list(...)
-        if (length(dots) > 0L)
-            dots <- dots[names(dots) != "exact"]
-        if (!missing(j) || length(dots) > 0L)
-            stop("incorrect number of subscripts")
-        .get_DelayedArray_element(x, i)
-    }
-)
-
-
 ### - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 ### Show
 ###
@@ -909,7 +735,7 @@ setMethod("show", "DelayedArray",
 
 ### Note that combining arrays with c() is NOT an endomorphism!
 setMethod("c", "DelayedArray",
-    function (x, ..., recursive=FALSE)
+    function(x, ..., recursive=FALSE)
     {
         if (!identical(recursive, FALSE))
             stop(wmsg("\"c\" method for DelayedArray objects ",
@@ -933,53 +759,3 @@ split.DelayedArray <- function(x, f, drop=FALSE, ...)
     splitAsList(x, f, drop=drop, ...)
 setMethod("split", c("DelayedArray", "ANY"), split.DelayedArray)
 
-
-### - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
-### Binding
-###
-### We only support binding DelayedArray objects along the rows or the cols
-### at the moment. No binding along an arbitrary dimension yet! (i.e. no
-### "abind" method yet)
-###
-
-### arbind() and acbind()
-
-.DelayedArray_arbind <- function(...) stash_DelayedAbind(..., along=1L)
-.DelayedArray_acbind <- function(...) stash_DelayedAbind(..., along=2L)
-
-setMethod("arbind", "DelayedArray", .DelayedArray_arbind)
-setMethod("acbind", "DelayedArray", .DelayedArray_acbind)
-
-### rbind() and cbind()
-
-setMethod("rbind", "DelayedMatrix", .DelayedArray_arbind)
-setMethod("cbind", "DelayedMatrix", .DelayedArray_acbind)
-
-.as_DelayedMatrix_objects <- function(objects)
-{
-    lapply(objects,
-        function(object) {
-            if (length(dim(object)) != 2L)
-                stop(wmsg("cbind() and rbind() are not supported on ",
-                          "DelayedArray objects that don't have exactly ",
-                          "2 dimensions. Please use acbind() or arbind() ",
-                          "instead."))
-            as(object, "DelayedMatrix")
-        })
-}
-
-.DelayedArray_rbind <- function(...)
-{
-    objects <- .as_DelayedMatrix_objects(list(...))
-    do.call("rbind", objects)
-}
-
-.DelayedArray_cbind <- function(...)
-{
-    objects <- .as_DelayedMatrix_objects(list(...))
-    do.call("cbind", objects)
-}
-
-setMethod("rbind", "DelayedArray", .DelayedArray_rbind)
-setMethod("cbind", "DelayedArray", .DelayedArray_cbind)
-
